"""Implementation of pairs for Python.

FIXME: not all the primitives are tail recursive!  Must fix this!
"""


from symbol import Symbol
from error import SchemeError
import unittest


## The NIL atom
NIL = []


class Cons(tuple): pass


def isNull(x):
    return x is NIL


def isPair(p):
    return type(p) is Cons


def schemeCons(head, rest):
    """Returns the concatentation of head with rest."""
    return Cons([head, rest])



def schemeCar(p):
    if not isPair(p):
        raise SchemeError, "CAR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    else:
        return p[0]


def schemeCdr(p):
    if not isPair(p):
        raise SchemeError, "CDR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    else:
        return p[1]


def schemeCddr(p):
    return schemeCdr(schemeCdr(p))

def schemeCdddr(p):
    return schemeCdr(schemeCdr(schemeCdr(p)))


def schemeCadr(p):
    return schemeCar(schemeCdr(p))

def schemeCaddr(p):
    return schemeCar(schemeCdr(schemeCdr(p)))

def schemeCadddr(p):
    return schemeCar(schemeCdr(schemeCdr(schemeCdr(p))))



## Hmm... some of these functions really belong in builtins.py.

def schemeSetCarBang(pair, element):
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[0] = element
    return Symbol("ok")


def schemeSetCdrBang(pair, element):
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[1] = element
    return Symbol("ok")


def isList(p):
    ## rewritten to avoid consuming stack
    while True:
        if isNull(p): return 1
        if not isPair(p): return 0
        p = schemeCdr(p)


def isDottedPair(p):
    if not isPair(p): return 0
    elif isPair(schemeCdr(p)) or isNull(schemeCdr(p)): return 0
    else: return 1


def schemeLength(p):
    length = 0
    while True:
        if isNull(p): return length
        length += 1
        p = schemeCdr(p)


def schemeReverse(p):
    """Reverses a list."""
    result = NIL
    while not isNull(p):
        result = schemeCons(schemeCar(p), result)
        p = schemeCdr(p)
    return result



def listMap(f, p):
    """Maps a function f across p."""
    resultsRev = NIL
    while not isNull(p):
        resultsRev = schemeCons(f(schemeCar(p)), resultsRev)
        p = schemeCdr(p)
    return schemeReverse(resultsRev)


def schemeCount(x, p):
    if isNull(p): return 0
    if x == schemeCar(p):
        return 1 + schemeCount(x, schemeCdr(p))
    else:
        return schemeCount(x, schemeCdr(p))

    

def appendTwo(front, back):
    if isNull(front): return back
    return schemeCons(schemeCar(front),
                      appendTwo(schemeCdr(front), back))
    


## Fixme: not tail recursive!
def schemeAppend(*lists):
    appended_list = lists[-1]
    for i in xrange(len(lists)-2, -1, -1):
        next_list = lists[i]
        appended_list = appendTwo(next_list, appended_list)
    return appended_list



def pairsToPythonList(pair):
    """Does a shallow conversion of a pair chain to a Python list."""
    elements = []
    while not isNull(pair):
        elements.append(schemeCar(pair))
        pair = schemeCdr(pair)
    return elements


def schemeList(*elements):
    """Does a shallow conversion of a Python list to a pair chain."""
    result = list(elements)
    result.reverse()
    return reduce(lambda x, y: schemeCons(y, x), [NIL] + result)


class PairTests(unittest.TestCase):
    def testNull(self):
        self.assert_(isList(NIL))

    def testReversal(self):
        self.assertEquals(schemeList(1, 2, 3, 4, 5),
                          schemeReverse(schemeList(5, 4, 3, 2, 1)))

    def testMapping(self):
        self.assertEquals(schemeList(2, 4, 6, 8),
                          listMap(lambda x: x*2, schemeList(1, 2, 3, 4)))



if __name__ == '__main__':
    unittest.main()
